package com.sillycycle.bagleyd.skewb;

/*
 * @(#)SkewbSolve.java
 *
 * Taken from the Mickey's Challenge by Christoph Bandelow
 *
 * Copyright 2009 - 2018  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>SkewbSolve</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class SkewbSolve {
	private static JFrame frame;
	private static Skewb skewb;
	private static Random generator;
	private static final boolean DEBUG = false;
	//private static int counter; // debugging

	SkewbSolve(JFrame frame, Skewb skewb) {
		SkewbSolve.frame = frame;
		SkewbSolve.skewb = skewb;
		generator = new Random(System.nanoTime());
		//this.counter = 0;
	}

	boolean moveAPiece(int face, int position, int direction,
			int control) {
		if (!getFlag()) {
			return false;
		}
		skewb.movePuzzleDelay(face, position, direction,
			control, Skewb.DOUBLE);
		return true;
	}

	static int findSquare(int color) {
		int i;

		for (i = 0; i < Skewb.MAX_FACES; i++) {
			if (skewb.cubeLoc[i][Skewb.MAX_ORIENT].face == color)
				return i;
		}
		return Skewb.MAX_FACES;
	}

	static boolean fixedCorner(int corner)
	{
		return ((((corner & 4) / 4) + ((corner & 2) / 2) +
			(corner & 1)) % 2 == 0);
	}

	static final int MAX_CORNERS = 8;

	static final int skewbCorner[][]/*[MAX_CORNERS=8][3]*/ =
	{
		{1, 4, 5},
		{1, 2, 4},
		{0, 1, 5},
		{0, 2, 1},
		{3, 5, 4},
		{2, 3, 4},
		{0, 5, 3},
		{0, 3, 2}
	};

	static final int skewbCornerOrient[][]/*[MAX_CORNERS][3]*/ =
	{
		{2, 2, 3},
		{1, 2, 3},
		{3, 3, 2},
		{2, 3, 0},
		{1, 0, 1},
		{1, 2, 0},
		{0, 1, 0},
		{1, 3, 0}
	};

	/* Corner layout, bits correspond to position
	   2 6
	 2 3 7 6
	 0 1 5 4
	   0 4
	   2 6
	Bandelow uses a different notation, advantage with his notation is that
	odd and even corners do not exchange.  With my notation, 2 bits and only
	2 bits are different to exchange.
	 */

	static int cornerColor(int corner, int n) {
		return skewb.cubeLoc[skewbCorner[corner][n]][skewbCornerOrient[corner][n]].face;
	}

	static int findCorner(int corner) {
		/* only need to check 4 corners */
		/* -0-, _3_, 5, 6 or 1, 2, -4-, _7_ */
		int i, j, t, count;
		int[] c = new int[3];

		c[0] = skewbCorner[corner][0];
		c[1] = skewbCorner[corner][1];
		c[2] = skewbCorner[corner][2];
		for (i = 0; i < MAX_CORNERS; i++) {
			/* fixed and free corners cannot interchange */
			if ((fixedCorner(corner) && (fixedCorner(i))) ||
					(!fixedCorner(corner) && (!fixedCorner(i)))) {
				count = 0;
				for (j = 0; j < 3; j++) {
					t = cornerColor(i, j);
					if (c[0] == t || c[1] == t || c[2] == t)
						count++;
				}
				if (count == 3) /* Got all 3 */
					return i;
			}
		}
		return MAX_CORNERS;
	}

	void setFirstSquare() {
		int i = findSquare(0); /* Red */

		if (DEBUG)
			System.out.println("firstSquare " + i);
		switch (i) {
		default:
		case 0:
			if (skewb.orientChoice &&
					skewb.cubeLoc[0][Skewb.MAX_ORIENT].rotation == 2) {
				if (generator.nextInt(2) == 1) {
					moveAPiece(0, 3, Skewb.BL, 1);
					moveAPiece(0, 0, Skewb.TL, 1);
				} else {
					moveAPiece(0, 0, Skewb.BR, 1);
					moveAPiece(0, 3, Skewb.TR, 1);
				}
			}
			break;
		case 1:
			if (skewb.cubeLoc[1][Skewb.MAX_ORIENT].rotation == 0) {
				moveAPiece(0, 1, Skewb.TR, 1);
			} else {
				moveAPiece(0, 0, Skewb.BR, 1);
			}
			break;
		case 2:
			if (skewb.cubeLoc[2][Skewb.MAX_ORIENT].rotation == 1) {
				moveAPiece(0, 0, Skewb.TL, 1);
			} else {
				moveAPiece(0, 3, Skewb.TR, 1);
			}
			break;
		case 3:
			if (skewb.cubeLoc[3][Skewb.MAX_ORIENT].rotation == 0) {
				moveAPiece(0, 2, Skewb.TL, 1);
			} else {
				moveAPiece(0, 3, Skewb.BL, 1);
			}
			break;
		case 4:
			if (skewb.cubeLoc[4][Skewb.MAX_ORIENT].rotation == 0) {
				if (generator.nextInt(2) == 1) {
					moveAPiece(2, 3, Skewb.BL, 1);
					moveAPiece( 0, 3, Skewb.BL, 1);
				} else {
					moveAPiece(2, 0, Skewb.BR, 1);
					moveAPiece(0, 0, Skewb.BR, 1);
				}
			} else {
				if (generator.nextInt(2) == 1) {
					moveAPiece(2, 3, Skewb.TR, 1);
					moveAPiece(0, 3, Skewb.TR, 1);
				} else {
					moveAPiece(2, 0, Skewb.TL, 1);
					moveAPiece(0, 0, Skewb.TL, 1);
				}
			}
			break;
		case 5:
			if ( skewb.cubeLoc[5][Skewb.MAX_ORIENT].rotation == 1) {
				moveAPiece(0, 2, Skewb.BR, 1);
			} else {
				moveAPiece(0, 1, Skewb.BL, 1);
			}
			break;
		}
		if (skewb.cubeLoc[0][Skewb.MAX_ORIENT].rotation == 3) {
			moveAPiece(2, 3, Skewb.LEFT, 2);
		} else if (skewb.cubeLoc[0][Skewb.MAX_ORIENT].rotation == 1) {
			moveAPiece(2, 3, Skewb.RIGHT, 2);
		}
	}

	void setSecondSquare() {
		int i = findSquare(2); /* White */

		if (DEBUG)
			System.out.println("secondSquare " + i);
		switch (i) {
		default:
		case 1:
			if (skewb.orientChoice &&
					skewb.cubeLoc[1][Skewb.MAX_ORIENT].rotation == 1) {
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
			} else {
				moveAPiece(0, 2, Skewb.BR, 0);
			}
			return;
		case 2:
			if (skewb.orientChoice &&
					skewb.cubeLoc[2][Skewb.MAX_ORIENT].rotation != 0) {
				if (generator.nextInt(2) == 1) {
					moveAPiece(0, 2, Skewb.BR, 0);
					moveAPiece(0, 1, Skewb.TR, 0);
				} else {
					moveAPiece(0, 1, Skewb.BL, 0);
					moveAPiece(0, 2, Skewb.TL, 0);
				}
			}
			return;
		case 3:
			if (skewb.orientChoice &&
					skewb.cubeLoc[3][Skewb.MAX_ORIENT].rotation != 1) {
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
			} else {
				moveAPiece(0, 1, Skewb.BL, 0);
			}
			return;
		case 4:
			if (skewb.cubeLoc[4][Skewb.MAX_ORIENT].rotation == 1)
				moveAPiece(0, 2, Skewb.TL, 0);
			else
				moveAPiece(0, 1, Skewb.TR, 0);
			return;
		case 5:
			if (generator.nextInt(2) == 1)
				moveAPiece(0, 0, Skewb.TL, 0);
			else
				moveAPiece(0, 3, Skewb.TR, 0);
			if (skewb.cubeLoc[4][Skewb.MAX_ORIENT].rotation == 1)
				moveAPiece(0, 2, Skewb.TL, 0);
			else
				moveAPiece(0, 1, Skewb.TR, 0);
			return;
		case 0:
			System.out.println("Wrong face " + i);
		}
	}

	void setFirstCorner() {
		/* 0, 3, 5, 6 */
		int i = findCorner(3); /* Red, White, Blue */

		if (DEBUG)
			System.out.println("firstCorner " + i);
		/* check if correctly placed already */
		if (i == 3) {
			if (skewb.cubeLoc[1][0].face == 0) {
				/* Diagram 4, turns 3 CW */
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
			} else if (skewb.cubeLoc[2][3].face == 0) {
				/* Reverse Diagram 4, turns 3 CCW */
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
			/* } else if (skewb.cubeLoc[0][2].face == 0) { */
			}
			return;
		}
		if (i != 0) {
			/* move to 0 if not already */
			/*if (i == 3) {
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
			} else*/ if (i == 5) {
				moveAPiece(0, 0, Skewb.BR, 0);
			} else if (i == 6) {
				moveAPiece(0, 0, Skewb.TL, 0);
			}
		}
		/* rotate "turnstile" if needed */
		if (skewb.cubeLoc[1][2].face == 1) {
			if (generator.nextInt(2) == 1) {
				moveAPiece(0, 3, Skewb.BL, 0);
			} else {
				moveAPiece(0, 3, Skewb.TR, 0);
			}
		}
		if (skewb.cubeLoc[1][2].face == 0) {
			/* Diagram 3b */
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 0, Skewb.BR, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
			return;
		}
		/* Diagram 3a */
		moveAPiece(0, 1, Skewb.TR, 0);
		moveAPiece(0, 2, Skewb.BR, 0);
		moveAPiece(0, 1, Skewb.BL, 0);
	}

	void setSecondCorner() {
		/* 1, 2, 4, 7 */
		int i = findCorner(7); /* Red, Green, White */

		if (DEBUG)
			System.out.println("secondCorner " + i);
		/* check if correctly placed already */
		if (i == 7) {
			if (skewb.cubeLoc[3][3].face == 0) {
				/* Mirror Diagram 4, turns 3 CCW */
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
			} else if (skewb.cubeLoc[2][0].face == 0) {
				/* Mirror Reverse Diagram 4, turns 3 CW */
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
			/* } else if (skewb.cubeLoc[0][1].face == 0) { */
			}
			return;
		}
		if (i != 4) {
			/* move to 4 if not already */
			/*if (i == 7) {
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
			} else*/ if (i == 1) {
				moveAPiece(0, 3, Skewb.BL, 0);
			} else if (i == 2) {
				moveAPiece(0, 3, Skewb.TR, 0);
			}
		}
		/* rotate "turnstile" if needed */
		if (skewb.cubeLoc[3][1].face == 3) {
			if (generator.nextInt(2) == 1) {
				moveAPiece(0, 0, Skewb.BR, 0);
			} else {
				moveAPiece(0, 0, Skewb.TL, 0);
			}
		}
		if (skewb.cubeLoc[3][1].face == 0) {
			/* Mirror Diagram 3b */
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 3, Skewb.BL, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
		} else { /* Mirror Diagram 3a */
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
		}
	}

	void otherSide() {
		if (generator.nextInt(2) == 1) {
			moveAPiece(0, 2, Skewb.BR, 1);
			moveAPiece(0, 0, Skewb.BR, 1);
		} else {
			moveAPiece(0, 1, Skewb.BL, 1);
			moveAPiece(0, 3, Skewb.BL, 1);
		}
	}

	void setThirdSquare() {
		int i = findSquare(4); /* Pink */

		if (DEBUG) {
			System.out.print("thirdSquare " + i + ", ");
			if (i == Skewb.MAX_FACES)
				System.out.println("i == MAX_FACES");
			else
				System.out.println(skewb.cubeLoc[i][Skewb.MAX_ORIENT].face + " " +
					skewb.cubeLoc[i][Skewb.MAX_ORIENT].rotation);
		}
		switch (i) {
		default:
		case 0:
			if (skewb.orientChoice &&
					skewb.cubeLoc[0][Skewb.MAX_ORIENT].rotation != 0) {
				if (generator.nextInt(2) == 1) {
					moveAPiece(3, 3, Skewb.BL, 0);
					moveAPiece(1, 0, Skewb.TL, 0);
				} else {
					moveAPiece(1, 0, Skewb.BR, 0);
					moveAPiece(3, 3, Skewb.TR, 0);
				}
			}
			return;
		case 1:
			if (skewb.orientChoice &&
					skewb.cubeLoc[1][Skewb.MAX_ORIENT].rotation == 0) {
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
			} else {
				moveAPiece(3, 3, Skewb.BL, 0);
			}
			return;
		case 2:
			if (skewb.cubeLoc[2][Skewb.MAX_ORIENT].rotation == 1) {
				moveAPiece(3, 3, Skewb.TR, 0);
			} else {
				moveAPiece(1, 0, Skewb.TL, 0);
			}
			return;
		case 3:
			if (skewb.orientChoice &&
					skewb.cubeLoc[3][Skewb.MAX_ORIENT].rotation == 0) {
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
			} else {
				moveAPiece(1, 0, Skewb.BR, 0);
			}
			return;
		case 4:
		case 5:
			System.out.println("Wrong face " + i);
		}
	}

	void placeThirdCorner() {
		/* 0, 3, 5, 6 */
		int i = findCorner(0); /* Pink, Yellow, Blue */

		if (DEBUG)
			System.out.println("thirdCorner " + i);
		/* check if correctly placed already */
		if (i == 3) {
			return;
		}
		/* move to 3 if not already */
		switch (i) {
		case 5:
			/* Diagram 6b, 7b */
			if (generator.nextInt(2) == 1) {
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
			} else {
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
			}
			break;
		case 6:
			/* Diagram 6a, 7a */
			if (generator.nextInt(2) == 1) {
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
			} else {
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
			}
			break;
		default:
			System.out.println("Wrong corner " + i);
		}
	}

	void placeFourthCorner() {
		/* 1, 2, 4, 7 */
		int i = findCorner(1); /* Pink, Green, Yellow */

		if (DEBUG)
			System.out.println("forthCorner " + i);
		/* check if correctly placed already */
		if (i == 2) {
			return;
		}
		switch (i) {
		case 1:
			/* Diagram 6c, 7c */
			if (generator.nextInt(2) == 1) {
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
			} else {
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
			}
			break;
		case 7:
			/* Diagram 6d, 7d */
			if (generator.nextInt(2) == 1) {
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
			} else {
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
			}
			break;
		default:
			System.out.println("Wrong corner " + i);
		}
	}

	/* This does not have to be solved for Mickey's challenge,
	   but needed on the skewb. */
	/* Its more efficient to check more things and use all the moves
	   available from "Mickey's Challenge" book, but going to cut down
	   for now. */
	void setFifthSixthCorner() {
		/* 3, 7 */
		int i = skewb.cubeLoc[0][3].face;
		int j = skewb.cubeLoc[0][0].face;
		int ri, rj;

		switch (i) {
		case 1:
			ri = 1;
			break;
		case 2:
			ri = -1;
			break;
		case 4:
			ri = 0;
			break;
		default:
			System.out.println("Wrong corner 3");
			return;
		}
		switch (j) {
		case 2:
			rj = 1;
			break;
		case 3:
			rj = -1;
			break;
		case 4:
			rj = 0;
			break;
		default:
			System.out.println("Wrong corner 7");
			return;
		}
		if (DEBUG) {
			System.out.print("fifthSixthCorner face 4");
			System.out.print(": 3 " + i);
			System.out.print(", 7 " + j);
			System.out.println(",  " + ri + " " + rj);
		}
		if (ri == 0 && rj == 0)
			return;
		if (ri == 1 && rj == 1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Triangle twister upsidedown, TT2 */
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
			} else {
				/* Diagram 11, Triangle twister upsidedown, TT2 */
				moveAPiece(2, 3, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.TR, 0);
				moveAPiece(2, 0, Skewb.TL, 0);
				moveAPiece(2, 3, Skewb.TR, 0);
				moveAPiece(2, 0, Skewb.TL, 0);
			}
		} else if (ri == -1 && rj == -1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Reverse Triangle twister upsidedown, TT2 */
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
			} else {
				/* Diagram 11, Reverse Mirror Triangle twister, TT2 */
				moveAPiece(2, 0, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.TL, 0);
				moveAPiece(2, 3, Skewb.TR, 0);
				moveAPiece(2, 0, Skewb.TL, 0);
				moveAPiece(2, 3, Skewb.TR, 0);
			}
		} else if (ri == -1 && rj == 0) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 8, Reverse Mirror Triangle twister, TT */
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
			} else {
				/* Diagram 10, Reverse Mirror Triangle twister, TT */
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
			}
		} else if (rj == -1 && ri == 0) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 8, Triangle twister, TT */
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
			} else {
				/* Diagram 10, Triangle twister, TT */
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
			}
		} else if (ri == 1 && rj == 0) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 8, Mirror Triangle twister, TT */
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
			} else {
				/* Diagram 10, Mirror Triangle twister, TT */
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
			}
		} else if (rj == 1 && ri == 0) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 8, Reverse Triangle twister, TT */
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
			} else {
				/* Diagram 10, Reverse Triangle twister, TT */
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
			}
		} else if (rj == -1 && ri == 1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Triangle twister 90 degrees, TT2 */
				moveAPiece(5, 2, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.TR, 0);
				moveAPiece(5, 2, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.TR, 0);
				moveAPiece(5, 2, Skewb.TL, 0);
				moveAPiece(2, 3, Skewb.BL, 0);
				moveAPiece(5, 2, Skewb.TL, 0);
				moveAPiece(2, 3, Skewb.BL, 0);
			} else {
				/* Diagram 11, Triangle twister 90 degrees, TT2 */
				moveAPiece(5, 1, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.TL, 0);
				moveAPiece(5, 1, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.TL, 0);
				moveAPiece(5, 1, Skewb.TR, 0);
				moveAPiece(2, 0, Skewb.BR, 0);
				moveAPiece(5, 1, Skewb.TR, 0);
				moveAPiece(2, 0, Skewb.BR, 0);
			}
		} else if (rj == 1 && ri == -1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Triangle twister 270 degrees, TT2 */
				moveAPiece(2, 3, Skewb.TR, 0);
				moveAPiece(5, 2, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.TR, 0);
				moveAPiece(5, 2, Skewb.BR, 0);
				moveAPiece(2, 3, Skewb.BL, 0);
				moveAPiece(5, 2, Skewb.TL, 0);
				moveAPiece(2, 3, Skewb.BL, 0);
				moveAPiece(5, 2, Skewb.TL, 0);
			} else {
				/* Diagram 11, Triangle twister 270 degrees, TT2 */
				moveAPiece(2, 0, Skewb.TL, 0);
				moveAPiece(5, 1, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.TL, 0);
				moveAPiece(5, 1, Skewb.BL, 0);
				moveAPiece(2, 0, Skewb.BR, 0);
				moveAPiece(5, 1, Skewb.TR, 0);
				moveAPiece(2, 0, Skewb.BR, 0);
				moveAPiece(5, 1, Skewb.TR, 0);
			}
		}
	}

	void placeRemainingSquares() {
		int i = findSquare(5); /* 1, 2, 3 */
		int j;

		if (DEBUG) {
			System.out.print("remainingSquares " + i + ", ");
			if (i == Skewb.MAX_FACES)
				System.out.println("i == MAX_FACES");
			else
				System.out.println(skewb.cubeLoc[i][Skewb.MAX_ORIENT].face + " " +
					skewb.cubeLoc[i][Skewb.MAX_ORIENT].rotation);
		}
		switch (i) {
		case 1:
			/* Diagram 12, Square Switcher, SS */
			for (j = 0; j < 2; j++) {
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
			}
			return;
		case 2:
			return;
		case 3:
			/* Diagram 12, Reverse Square Switcher, SS */
			for (j = 0; j < 2; j++) {
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
			}
			return;
		default:
		case 0:
		case 4:
		case 5:
			System.out.println("Wrong face " + i);
		}
	}

	void setRemainingCorners() {
		/* 1, 5, 0, 4 */
		int[] f = new int[4], r = new int[4];

		f[0] = skewb.cubeLoc[2][0].face;
		f[1] = skewb.cubeLoc[2][1].face;
		f[2] = skewb.cubeLoc[2][2].face;
		f[3] = skewb.cubeLoc[2][3].face;
		if (DEBUG) {
			System.out.print("remainingCorners face 5");
			System.out.print(": 5 " + f[0]);
			System.out.print(", 4 " + f[1]);
			System.out.print(", 0 " + f[2]);
			System.out.println(", 1 " + f[3]);
		}
		switch (f[0]) {
		case 3:
			r[0] = -1;
			break;
		case 4:
			r[0] = 1;
			break;
		case 5:
			r[0] = 0;
			break;
		default:
			System.out.println("Wrong corner 5");
			return;
		}
		switch (f[1]) {
		case 0:
			r[1] = -1;
			break;
		case 3:
			r[1] = 1;
			break;
		case 5:
			r[1] = 0;
			break;
		default:
			System.out.println("Wrong corner 4");
			return;
		}
		switch (f[2]) {
		case 0:
			r[2] = 1;
			break;
		case 1:
			r[2] = -1;
			break;
		case 5:
			r[2] = 0;
			break;
		default:
			System.out.println("Wrong corner 0");
			return;
		}
		switch (f[3]) {
		case 1:
			r[3] = 1;
			break;
		case 4:
			r[3] = -1;
			break;
		case 5:
			r[3] = 0;
			break;
		default:
			System.out.println("Wrong corner 1");
			return;
		}
		if (DEBUG) {
			System.out.println("remainingCorners " +
				f[0] + " " + f[1] + " " + f[2] + " " + f[3] + " " +
				r[0] + " " + r[1] + " " + r[2] + " " + r[3]);
		}
		if (r[0] == 0 && r[1] == 0 && r[2] == 0)
			return;
		if (r[0] == 0 && r[2] == 0) {
			if (r[1] == 1) {
				/* Diagram 13, 90 degrees, mirror */
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
			} else {
				/* Diagram 13, 90 degrees, mirror upsidedown */
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
			}
			return;
		}
		if (r[1] == 0 && r[3] == 0) {
			if (r[0] == 1) {
				/* Diagram 13, 90 degrees */
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
			} else {
				/* Diagram 13, 90 degrees, upsidedown */
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 0, Skewb.TL, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 3, Skewb.BL, 0);
				moveAPiece(0, 0, Skewb.BR, 0);
				moveAPiece(0, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
			}
			return;
		}
		if (r[0] == 1 && r[1] == 1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Triangle twister TT2 90 degrees */
				moveAPiece(4, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(4, 0, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(4, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(4, 0, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
			} else {
				/* Diagram 11, Triangle twister TT2 90 degrees */
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(3, 2, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(3, 2, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(3, 2, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(3, 2, Skewb.TL, 0);
			}
			return;
		}
		if (r[1] == 1 && r[2] == 1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Triangle twister TT2, 180 degrees */
				moveAPiece(1, 1, Skewb.TR, 0);
				moveAPiece(3, 2, Skewb.TL, 0);
				moveAPiece(1, 1, Skewb.TR, 0);
				moveAPiece(3, 2, Skewb.TL, 0);
				moveAPiece(1, 1, Skewb.BL, 0);
				moveAPiece(3, 2, Skewb.BR, 0);
				moveAPiece(1, 1, Skewb.BL, 0);
				moveAPiece(3, 2, Skewb.BR, 0);
			} else {
				/* Diagram 11, Triangle twister TT2 180 degrees */
				moveAPiece(4, 0, Skewb.BR, 0);
				moveAPiece(4, 3, Skewb.BL, 0);
				moveAPiece(4, 0, Skewb.BR, 0);
				moveAPiece(4, 3, Skewb.BL, 0);
				moveAPiece(4, 0, Skewb.TL, 0);
				moveAPiece(4, 3, Skewb.TR, 0);
				moveAPiece(4, 0, Skewb.TL, 0);
				moveAPiece(4, 3, Skewb.TR, 0);
			}
			return;
		}
		if (r[2] == 1 && r[3] == 1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Triangle twister TT2, 270 degrees */
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(4, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(4, 3, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(4, 3, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(4, 3, Skewb.BL, 0);
			} else {
				/* Diagram 11, Triangle twister TT2 270 degrees */
				moveAPiece(1, 1, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(1, 1, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(1, 1, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(1, 1, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
			}
			return;
		}
		if (r[3] == 1 && r[0] == 1) {
			if (generator.nextInt(2) == 1) {
				/* Diagram 9, Triangle twister TT2 */
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.BL, 0);
				moveAPiece(1, 0, Skewb.BR, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
				moveAPiece(3, 3, Skewb.TR, 0);
				moveAPiece(1, 0, Skewb.TL, 0);
			} else {
				/* Diagram 11, Triangle twister TT2 */
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.TL, 0);
				moveAPiece(0, 1, Skewb.TR, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
				moveAPiece(0, 2, Skewb.BR, 0);
				moveAPiece(0, 1, Skewb.BL, 0);
			}
			return;
		}
		System.out.println("Wrong corners");
	}

	void setRemainingSquares() {
		int[] r = new int[2]; /* 1, 2, 3 */ /* third value inferred */

		r[0] = 2 - skewb.cubeLoc[1][Skewb.MAX_ORIENT].rotation;
		r[1] = skewb.cubeLoc[2][Skewb.MAX_ORIENT].rotation;

		if (DEBUG)
			System.out.println("remainingSquares " +
				r[0] + " " + r[1]);
		if (r[0] == 0 && r[1] == 0)
			return;
		if (r[0] == 0) {
			/* Diagram 16, Twists 2 squares */
			moveAPiece(0, 3, Skewb.TR, 0);
			moveAPiece(0, 0, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
			moveAPiece(0, 3, Skewb.BL, 0);
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 0, Skewb.BR, 0);
			moveAPiece(0, 3, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
			moveAPiece(0, 3, Skewb.BL, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
		} else if (r[1] == 0) {
			/* Diagram 17, Twists 2 opposite squares */
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
			moveAPiece(0, 0, Skewb.TL, 0);
			moveAPiece(0, 3, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
			moveAPiece(0, 0, Skewb.BR, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 3, Skewb.BL, 0);
			moveAPiece(0, 0, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
			moveAPiece(0, 0, Skewb.BR, 0);
		} else {
			/* Diagram 16, Mirror Twists 2 squares */
			moveAPiece(0, 0, Skewb.TL, 0);
			moveAPiece(0, 3, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
			moveAPiece(0, 0, Skewb.BR, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 3, Skewb.BL, 0);
			moveAPiece(0, 0, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
			moveAPiece(0, 0, Skewb.BR, 0);
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.BR, 0);
		}
	}

/*
	void diagram13() {
		//Twister of 2 squares and 2 corners
		moveAPiece(0, 1, Skewb.TR, 0);
		moveAPiece(0, 2, Skewb.BR, 0);
		moveAPiece(0, 3, Skewb.BL, 0);
		moveAPiece(0, 2, Skewb.TL, 0);
		moveAPiece(0, 0, Skewb.BR, 0);
		moveAPiece(0, 1, Skewb.BL, 0);
		moveAPiece(0, 3, Skewb.TR, 0);
		moveAPiece(0, 1, Skewb.BL, 0);
		moveAPiece(0, 3, Skewb.BL, 0);
		moveAPiece(0, 0, Skewb.TL, 0);
	}

	void diagram14() {
		//Twists 2 squares, simple but Diagram 16 more efficient
		int i;

		for (i = 0; i < 6; i++) {
			moveAPiece(0, 3, Skewb.TR, 0);
			moveAPiece(0, 1, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
		}
	}

	void diagram15() {
		// Twister of 2 squares and 3 corners
		moveAPiece(0, 0, Skewb.BR, 0);
		moveAPiece(0, 2, Skewb.TL, 0);
		moveAPiece(0, 1, Skewb.BL, 0);
		moveAPiece(0, 0, Skewb.BR, 0);
		moveAPiece(0, 1, Skewb.BL, 0);
		moveAPiece(0, 2, Skewb.TL, 0);
		moveAPiece(0, 1, Skewb.BL, 0);
		moveAPiece(0, 0, Skewb.BR, 0);
		moveAPiece(0, 2, Skewb.TL, 0);
	}

	void diagram18() {
		// Twists 6 corners clockwise
		int i;

		for (i = 0; i < 3; i++) {
			moveAPiece(0, 3, Skewb.TR, 0);
			moveAPiece(0, 2, Skewb.TL, 0);
			moveAPiece(0, 1, Skewb.BL, 0);
			moveAPiece(0, 0, Skewb.BR, 0);
		}
	}
*/

	/* This procedure coordinates the solution process. */
	void solvePieces() {
		if (!skewb.checkSolved()) {
			setFirstSquare();
			setSecondSquare();
			if (generator.nextInt(2) == 1) {
				setFirstCorner();
				setSecondCorner();
			} else {
				setSecondCorner();
				setFirstCorner();
			}
			otherSide();
			setThirdSquare();
			if (generator.nextInt(2) == 1) {
				placeThirdCorner();
				placeFourthCorner();
			} else {
				placeFourthCorner();
				placeThirdCorner();
			}
			setFifthSixthCorner();
			placeRemainingSquares();
			setRemainingCorners();
			if (skewb.orientChoice)
				setRemainingSquares();
			otherSide();
		}
		((SkewbFrame) frame).callback(
			SkewbInterface.ACTION_COMPUTED, skewb.dim);
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solvePieces();
					runFlag = false;
				}
			}
		}
	}

	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
